import unittest

from wikibaseintegrator import WikibaseIntegrator
from wikibaseintegrator.wbi_config import config as wbi_config

wbi_config['USER_AGENT'] = 'WikibaseIntegrator-pytest/1.0 (test_entity_mediainfo.py)'

wbi = WikibaseIntegrator()


class TestEntityMediaInfo(unittest.TestCase):

    def test_get(self):
        # Test with complete id
        assert wbi.mediainfo.get('M75908279', mediawiki_api_url='https://commons.wikimedia.org/w/api.php').id == 'M75908279'
        # Test with numeric id as string
        assert wbi.mediainfo.get('75908279', mediawiki_api_url='https://commons.wikimedia.org/w/api.php').id == 'M75908279'
        # Test with numeric id as int
        assert wbi.mediainfo.get(75908279, mediawiki_api_url='https://commons.wikimedia.org/w/api.php').id == 'M75908279'

        # Test with invalid id
        with self.assertRaises(ValueError):
            wbi.mediainfo.get('L5')

        # Test with zero id
        with self.assertRaises(ValueError):
            wbi.mediainfo.get(0)

        # Test with negative id
        with self.assertRaises(ValueError):
            wbi.mediainfo.get(-1)

    def test_get_json(self):
        assert wbi.mediainfo.get('M75908279', mediawiki_api_url='https://commons.wikimedia.org/w/api.php').get_json()



from unittest import TestCase

from wikibaseintegrator.wbi_exceptions import ModificationFailed, SaveFailed, SearchError


class TestWbiExceptions(TestCase):
    @staticmethod
    def test_modification_failed():
        error_dict = {'error': {'*': 'See https://test.wikidata.org/w/api.php for API usage. '
                                     'Subscribe to the mediawiki-api-announce mailing list at '
                                     '&lt;https://lists.wikimedia.org/postorius/lists/mediawiki-api-announce.lists.wikimedia.org/&gt; '
                                     'for notice of API deprecations and breaking changes.',
                                'code': 'modification-failed',
                                'info': 'Item [[Q582|Q582]] already has label "MODIFIED LABEL" '
                                        'associated with language code en, using the same '
                                        'description text.',
                                'messages': [{'html': {'*': 'Item <a href="/wiki/Q582" '
                                                            'title="Q582">Q582</a> already has '
                                                            'label "MODIFIED LABEL" associated with '
                                                            'language code en, using the same '
                                                            'description text.'},
                                              'name': 'wikibase-validator-label-with-description-conflict',
                                              'parameters': ['MODIFIED LABEL',
                                                             'en',
                                                             '[[Q582|Q582]]']}]},
                      'servedby': 'mw1375'}

        modification_failed = ModificationFailed(error_dict['error'])

        assert str(modification_failed) == "'Item [[Q582|Q582]] already has label \"MODIFIED LABEL\" associated with language code en, using the same description text.'"
        assert modification_failed.code == 'modification-failed'
        assert modification_failed.info == 'Item [[Q582|Q582]] already has label "MODIFIED LABEL" associated with language code en, using the same description text.'
        assert 'wikibase-validator-label-with-description-conflict' in modification_failed.messages_names
        assert 'Q582' in modification_failed.get_conflicting_entity_ids
        assert 'en' in modification_failed.get_languages

    def test_modification_failed_no_dict(self):
        error_dict = {}
        with self.assertRaises(KeyError):
            ModificationFailed(error_dict['error'])

    def test_modification_failed_no_message(self):
        error_dict = {'error': {'*': 'See https://test.wikidata.org/w/api.php for API usage. '
                                     'Subscribe to the mediawiki-api-announce mailing list at '
                                     '&lt;https://lists.wikimedia.org/postorius/lists/mediawiki-api-announce.lists.wikimedia.org/&gt; '
                                     'for notice of API deprecations and breaking changes.',
                                'code': 'modification-failed',
                                'info': 'Item [[Q582|Q582]] already has label "MODIFIED LABEL" '
                                        'associated with language code en, using the same '
                                        'description text.'
                                },
                      'servedby': 'mw1375'}

        with self.assertRaises(KeyError):
            ModificationFailed(error_dict['error'])

    def test_failed_save_no_conflict(self):
        error_dict = {'error': {'*': 'See https://test.wikidata.org/w/api.php for API usage. '
                                     'Subscribe to the mediawiki-api-announce mailing list at '
                                     '&lt;https://lists.wikimedia.org/postorius/lists/mediawiki-api-announce.lists.wikimedia.org/&gt; '
                                     'for notice of API deprecations and breaking changes.',
                                'code': 'failed-save',
                                'info': 'The save has failed.',
                                'messages': [{'html': {'*': 'The save has failed.'},
                                              'name': 'wikibase-api-failed-save',
                                              'parameters': []}]},
                      'servedby': 'mw1425'}

        failed_save = SaveFailed(error_dict['error'])

        assert failed_save.get_conflicting_entity_ids == []

    def test_modification_failed_no_parameters(self):
        error_dict = {'error': {'*': 'See https://test.wikidata.org/w/api.php for API usage. '
                                     'Subscribe to the mediawiki-api-announce mailing list at '
                                     '&lt;https://lists.wikimedia.org/postorius/lists/mediawiki-api-announce.lists.wikimedia.org/&gt; '
                                     'for notice of API deprecations and breaking changes.',
                                'code': 'modification-failed',
                                'info': 'Item [[Q582|Q582]] already has label "MODIFIED LABEL" '
                                        'associated with language code en, using the same '
                                        'description text.',
                                'messages': [{'html': {'*': 'Item <a href="/wiki/Q582" '
                                                            'title="Q582">Q582</a> already has '
                                                            'label "MODIFIED LABEL" associated with '
                                                            'language code en, using the same '
                                                            'description text.'},
                                              'name': 'wikibase-validator-label-with-description-conflict',
                                              }]},
                      'servedby': 'mw1375'}

        modification_failed = ModificationFailed(error_dict['error'])
        with self.assertRaises(KeyError):
            _ = modification_failed.get_languages

    @staticmethod
    def test_failed_save():
        error_dict = {'error': {'*': 'See https://test.wikidata.org/w/api.php for API usage. '
                                     'Subscribe to the mediawiki-api-announce mailing list at '
                                     '&lt;https://lists.wikimedia.org/postorius/lists/mediawiki-api-announce.lists.wikimedia.org/&gt; '
                                     'for notice of API deprecations and breaking changes.',
                                'code': 'failed-save',
                                'info': 'The save has failed.',
                                'messages': [{'html': {'*': 'The save has failed.'},
                                              'name': 'wikibase-api-failed-save',
                                              'parameters': []},
                                             {'html': {'*': 'Property <a href="/wiki/Property:P50" '
                                                            'title="Property:P50">P50</a> already '
                                                            'has label "Depiction" associated with '
                                                            'language code en.'},
                                              'name': 'wikibase-validator-label-conflict',
                                              'parameters': ['Depiction',
                                                             'en',
                                                             '[[Property:P50|P50]]']},
                                             {'html': {'*': 'Property <a href="/wiki/Property:P50" '
                                                            'title="Property:P50">P50</a> already '
                                                            'has label "representación" associated '
                                                            'with language code es.'},
                                              'name': 'wikibase-validator-label-conflict',
                                              'parameters': ['representación',
                                                             'es',
                                                             '[[Property:P50|P50]]']}]},
                      'servedby': 'mw1425'}

        failed_save = SaveFailed(error_dict['error'])

        assert str(failed_save) == "'The save has failed.'"
        assert failed_save.code == 'failed-save'
        assert failed_save.info == 'The save has failed.'
        assert 'wikibase-api-failed-save' in failed_save.messages_names
        assert 'P50' in failed_save.get_conflicting_entity_ids
        assert len(failed_save.get_conflicting_entity_ids) == 1
        assert 'en' in failed_save.get_languages

    @staticmethod
    def test_searcherror():
        assert str(SearchError('SearchError')) == 'SearchError'



from __future__ import annotations

import logging
from copy import copy
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import ujson

from wikibaseintegrator import wbi_fastrun
from wikibaseintegrator.datatypes import BaseDataType
from wikibaseintegrator.models.claims import Claim, Claims
from wikibaseintegrator.wbi_enums import ActionIfExists
from wikibaseintegrator.wbi_exceptions import MissingEntityException, ModificationFailed, MWApiError
from wikibaseintegrator.wbi_helpers import delete_page, mediawiki_api_call_helper
from wikibaseintegrator.wbi_login import _Login

if TYPE_CHECKING:
    from wikibaseintegrator import WikibaseIntegrator

log = logging.getLogger(__name__)


class BaseEntity:
    ETYPE = 'base-entity'

    def __init__(self, api: 'WikibaseIntegrator' = None, title: str = None, pageid: int = None, lastrevid: int = None, type: str = None, id: str = None, claims: Claims = None,
                 is_bot: bool = None, login: _Login = None):
        if not api:
            from wikibaseintegrator import WikibaseIntegrator
            self.api = WikibaseIntegrator()
        else:
            self.api = copy(api)

        self.api.is_bot = is_bot or self.api.is_bot
        self.api.login = login or self.api.login

        self.title = title
        self.pageid = pageid
        self.lastrevid = lastrevid
        self.type = str(type or self.ETYPE)
        self.id = id
        self.claims = claims or Claims()

    @property
    def api(self) -> WikibaseIntegrator:
        return self.__api

    @api.setter
    def api(self, value: WikibaseIntegrator):
        from wikibaseintegrator import WikibaseIntegrator
        if not isinstance(value, WikibaseIntegrator):
            raise TypeError
        self.__api = value

    @property
    def title(self) -> Optional[str]:
        return self.__title

    @title.setter
    def title(self, value: Optional[str]):
        self.__title = value

    @property
    def pageid(self) -> Union[str, int, None]:
        return self.__pageid

    @pageid.setter
    def pageid(self, value: Union[str, int, None]):
        if isinstance(value, str):
            self.__pageid: Union[str, int, None] = int(value)
        else:
            self.__pageid = value

    @property
    def lastrevid(self) -> Optional[int]:
        return self.__lastrevid

    @lastrevid.setter
    def lastrevid(self, value: Optional[int]):
        self.__lastrevid = value

    @property
    def type(self) -> str:
        return self.__type

    @type.setter
    def type(self, value: str):
        self.__type = value

    @property
    def id(self) -> Optional[str]:
        return self.__id

    @id.setter
    def id(self, value: Optional[str]):
        self.__id = value

    @property
    def claims(self) -> Claims:
        return self.__claims

    @claims.setter
    def claims(self, value: Claims):
        if not isinstance(value, Claims):
            raise TypeError
        self.__claims = value

    def add_claims(self, claims: Union[Claim, List[Claim], Claims], action_if_exists: ActionIfExists = ActionIfExists.APPEND_OR_REPLACE) -> BaseEntity:
        """

        :param claims: A Claim, list of Claim or just a Claims object to add to this Claims object.
        :param action_if_exists: Replace or append the statement. You can force an addition if the declaration already exists.
            KEEP: The original claim will be kept and the new one will not be added (because there is already one with this property number)
            APPEND_OR_REPLACE: The new claim will be added only if the new one is different (by comparing values)
            FORCE_APPEND: The new claim will be added even if already exists
            REPLACE_ALL: The new claim will replace the old one
        :return: Return the updated entity object.
        """

        self.claims.add(claims=claims, action_if_exists=action_if_exists)

        return self

    def get_json(self) -> Dict[str, Union[str, Dict[str, List]]]:
        """
        To get the dict equivalent of the JSON representation of the entity.

        :return:
        """
        json_data: Dict = {
            'type': self.type,
            'claims': self.claims.get_json()
        }
        if self.id:
            json_data['id'] = self.id

        return json_data

    def from_json(self, json_data: Dict[str, Any]) -> BaseEntity:
        """
        Import a dictionary into BaseEntity attributes.

        :param json_data: A specific dictionary from MediaWiki API
        :return:
        """
        if 'missing' in json_data:  # TODO: 1.35 compatibility
            raise MissingEntityException('The MW API returned that the entity was missing.')

        if 'title' in json_data:  # TODO: 1.35 compatibility
            self.title = str(json_data['title'])
        if 'pageid' in json_data:  # TODO: 1.35 compatibility
            self.pageid = int(json_data['pageid'])
        self.lastrevid = int(json_data['lastrevid'])
        self.type = str(json_data['type'])
        self.id = str(json_data['id'])
        if 'claims' in json_data:  # 'claims' is named 'statements' in Wikimedia Commons MediaInfo
            self.claims = Claims().from_json(json_data['claims'])

        return self

    # noinspection PyMethodMayBeStatic
    def _get(self, entity_id: str, login: _Login = None, allow_anonymous: bool = True, is_bot: bool = None, **kwargs: Any) -> Dict:  # pylint: disable=no-self-use
        """
        Retrieve an entity in json representation from the Wikibase instance

        :param entity_id: The ID of the entity to retrieve
        :param login: A login instance
        :param allow_anonymous: Force a check if the query can be anonymous or not
        :param is_bot: Add the bot flag to the query
        :param kwargs: More arguments for Python requests
        :return: python complex dictionary representation of a json
        """

        params = {
            'action': 'wbgetentities',
            'ids': entity_id,
            'format': 'json'
        }

        login = login or self.api.login
        is_bot = is_bot if is_bot is not None else self.api.is_bot

        return mediawiki_api_call_helper(data=params, login=login, allow_anonymous=allow_anonymous, is_bot=is_bot, **kwargs)

    def clear(self, **kwargs: Any) -> Dict[str, Any]:
        """
        Use the `clear` parameter of `wbeditentity` API call to clear the content of the entity.
        The entity will be updated with an empty dictionary.

        :param kwargs: More arguments for _write() and Python requests
        :return: A dictionary representation of the edited Entity
        """
        return self._write(data={}, clear=True, **kwargs)

    def _write(self, data: Dict = None, summary: str = None, login: _Login = None, allow_anonymous: bool = False, clear: bool = False, is_bot: bool = None, **kwargs: Any) -> Dict[
        str, Any]:
        """
        Writes the entity JSON to the Wikibase instance and after successful write, returns the "entity" part of the response.

        :param data: The serialized object that is used as the data source. A newly created entity will be assigned an 'id'.
        :param summary: A summary of the edit
        :param login: A login instance
        :param allow_anonymous: Force a check if the query can be anonymous or not
        :param clear: If set, the complete entity is emptied before proceeding. The entity will not be saved before it is filled with the "data", possibly with parts excluded.
        :param is_bot: Add the bot flag to the query
        :param kwargs: More arguments for Python requests
        :return: A dictionary representation of the edited Entity
        """

        data = data or {}

        # if all_claims:
        #     data = json.JSONEncoder().encode(self.json_representation)
        # else:
        #     new_json_repr = {k: self.json_representation[k] for k in set(list(self.json_representation.keys())) - {'claims'}}
        #     new_json_repr['claims'] = {}
        #     for claim in self.json_representation['claims']:
        #         if [True for x in self.json_representation['claims'][claim] if 'id' not in x or 'remove' in x]:
        #             new_json_repr['claims'][claim] = copy.deepcopy(self.json_representation['claims'][claim])
        #             for statement in new_json_repr['claims'][claim]:
        #                 if 'id' in statement and 'remove' not in statement:
        #                     new_json_repr['claims'][claim].remove(statement)
        #             if not new_json_repr['claims'][claim]:
        #                 new_json_repr['claims'].pop(claim)
        #     data = json.JSONEncoder().encode(new_json_repr)

        payload: Dict[str, Any] = {
            'action': 'wbeditentity',
            'data': ujson.dumps(data),
            'format': 'json',
            'summary': summary
        }

        if not summary:
            payload.pop('summary')

        is_bot = is_bot if is_bot is not None else self.api.is_bot
        if is_bot:
            payload.update({'bot': ''})

        if clear:
            payload.update({'clear': True})

        if self.id:
            payload.update({'id': self.id})
        else:
            payload.update({'new': self.type})

        if self.lastrevid:
            payload.update({'baserevid': self.lastrevid})

        login = login or self.api.login

        try:
            json_result: dict = mediawiki_api_call_helper(data=payload, login=login, allow_anonymous=allow_anonymous, is_bot=is_bot, **kwargs)
        except Exception:
            logging.exception('Error while writing to the Wikibase instance')
            raise
        else:
            if 'error' in json_result:
                if 'code' in json_result['error'] and json_result['error']['code'] == 'modification-failed':
                    raise ModificationFailed(json_result['error'])

                raise MWApiError(json_result)

        return json_result['entity']

    def delete(self, login: _Login = None, allow_anonymous: bool = False, is_bot: bool = None, **kwargs: Any):
        """
        Delete the current entity. Use the pageid first if available and fallback to the page title.

        :param login: A wbi_login.Login instance
        :param allow_anonymous: Allow an unidentified edit to the MediaWiki API (default False)
        :param is_bot: Flag the edit as a bot
        :param reason: Reason for the deletion. If not set, an automatically generated reason will be used.
        :param deletetalk: Delete the talk page, if it exists.
        :param kwargs: Any additional keyword arguments to pass to mediawiki_api_call_helper and requests.request
        :return: The data returned by the API as a dictionary
        """

        login = login or self.api.login

        if not self.pageid and not self.title:
            raise ValueError("A pageid or a page title attribute must be set before deleting an entity object.")

        # If there is no pageid, fallback to using the page title. It's not the preferred method.
        if not self.pageid:
            return delete_page(title=self.title, pageid=None, login=login, allow_anonymous=allow_anonymous, is_bot=is_bot, **kwargs)
        else:
            if not isinstance(self.pageid, int):
                raise ValueError(f"The entity must have a pageid attribute correctly set ({self.pageid})")

            return delete_page(title=None, pageid=self.pageid, login=login, allow_anonymous=allow_anonymous, is_bot=is_bot, **kwargs)

    def write_required(self, base_filter: List[BaseDataType | List[BaseDataType]] = None, action_if_exists: ActionIfExists = ActionIfExists.REPLACE_ALL, **kwargs: Any) -> bool:
        fastrun_container = wbi_fastrun.get_fastrun_container(base_filter=base_filter, **kwargs)

        if base_filter is None:
            base_filter = []

        claims_to_check = []
        for claim in self.claims:
            if claim.mainsnak.property_number in base_filter:
                claims_to_check.append(claim)

        # TODO: Add check_language_data

        return fastrun_container.write_required(data=claims_to_check, cqid=self.id, action_if_exists=action_if_exists)

    def __repr__(self):
        """A mixin implementing a simple __repr__."""
        return "<{klass} @{id:x} {attrs}>".format(  # pylint: disable=consider-using-f-string
            klass=self.__class__.__name__,
            id=id(self) & 0xFFFFFF,
            attrs="\r\n\t ".join(f"{k}={v!r}" for k, v in self.__dict__.items()),
        )



import unittest

from wikibaseintegrator import WikibaseIntegrator
from wikibaseintegrator.wbi_config import config as wbi_config

wbi_config['USER_AGENT'] = 'WikibaseIntegrator-pytest/1.0 (test_entity_property.py)'

wbi = WikibaseIntegrator()


class TestEntityProperty(unittest.TestCase):

    def test_get(self):
        # Test with complete id
        assert wbi.property.get('P50').id == 'P50'
        # Test with numeric id as string
        assert wbi.property.get('50').id == 'P50'
        # Test with numeric id as int
        assert wbi.property.get(50).id == 'P50'

        # Test with invalid id
        with self.assertRaises(ValueError):
            wbi.property.get('L5')

        # Test with zero id
        with self.assertRaises(ValueError):
            wbi.property.get(0)

        # Test with negative id
        with self.assertRaises(ValueError):
            wbi.property.get(-1)

    def test_get_json(self):
        assert wbi.property.get('P50', mediawiki_api_url='https://commons.wikimedia.org/w/api.php').get_json()['labels']['fr']['value'] == 'auteur ou autrice'

    def test_create_property(self):
        wbi.property.new(datatype='wikibase-item')

    def test_long_item_id(self):
        assert wbi.property.get('Property:P582').id == 'P582'



import os
import sys
import unittest

import pytest

from wikibaseintegrator import wbi_login
from wikibaseintegrator.wbi_helpers import mediawiki_api_call_helper
# look for environment variables. if none set, don't do anything
from wikibaseintegrator.wbi_login import LoginError

WDUSER = os.getenv("WDUSER")
WDPASS = os.getenv("WDPASS")
OAUTH1_CONSUMER_TOKEN_NOT_OWNER_ONLY = os.getenv("OAUTH1_CONSUMER_TOKEN_NOT_OWNER_ONLY")
OAUTH1_CONSUMER_SECRET_NOT_OWNER_ONLY = os.getenv("OAUTH1_CONSUMER_SECRET_NOT_OWNER_ONLY")
OAUTH1_CONSUMER_TOKEN = os.getenv("OAUTH1_CONSUMER_TOKEN")
OAUTH1_CONSUMER_SECRET = os.getenv("OAUTH1_CONSUMER_SECRET")
OAUTH1_ACCESS_TOKEN = os.getenv("OAUTH1_ACCESS_TOKEN")
OAUTH1_ACCESS_SECRET = os.getenv("OAUTH1_ACCESS_SECRET")
OAUTH2_CONSUMER_TOKEN = os.getenv("OAUTH2_CONSUMER_TOKEN")
OAUTH2_CONSUMER_SECRET = os.getenv("OAUTH2_CONSUMER_SECRET")


def test_login():
    with unittest.TestCase().assertRaises(LoginError):
        login = wbi_login.Clientlogin(user='wrong', password='wrong')
        login.generate_edit_credentials()

    with unittest.TestCase().assertRaises(LoginError):
        login = wbi_login.Login(user='wrong', password='wrong')
        login.generate_edit_credentials()

    if WDUSER and WDPASS:
        assert wbi_login.Clientlogin(user=WDUSER, password=WDPASS)
        assert wbi_login.Login(user=WDUSER, password=WDPASS)
    else:
        print("no WDUSER or WDPASS found in environment variables", file=sys.stderr)


def test_oauth1():
    with unittest.TestCase().assertRaises(LoginError):
        login = wbi_login.OAuth1(consumer_token='wrong', consumer_secret='wrong')
        login.generate_edit_credentials()

    if OAUTH1_CONSUMER_TOKEN_NOT_OWNER_ONLY and OAUTH1_CONSUMER_SECRET_NOT_OWNER_ONLY:
        wbi_login.OAuth1(consumer_token=OAUTH1_CONSUMER_TOKEN_NOT_OWNER_ONLY, consumer_secret=OAUTH1_CONSUMER_SECRET_NOT_OWNER_ONLY)
    else:
        print("no OAUTH1_CONSUMER_TOKEN_NOT_OWNER_ONLY or OAUTH1_CONSUMER_SECRET_NOT_OWNER_ONLY found in environment variables", file=sys.stderr)


def test_oauth1_access():
    with unittest.TestCase().assertRaises(LoginError):
        login = wbi_login.OAuth1(consumer_token='wrong', consumer_secret='wrong', access_token='wrong', access_secret='wrong')
        login.generate_edit_credentials()

    if OAUTH1_CONSUMER_TOKEN and OAUTH1_CONSUMER_SECRET and OAUTH1_ACCESS_TOKEN and OAUTH1_ACCESS_SECRET:
        login = wbi_login.OAuth1(consumer_token=OAUTH1_CONSUMER_TOKEN, consumer_secret=OAUTH1_CONSUMER_SECRET, access_token=OAUTH1_ACCESS_TOKEN, access_secret=OAUTH1_ACCESS_SECRET)
        login.generate_edit_credentials()
    else:
        print("no OAUTH1_CONSUMER_TOKEN or OAUTH1_CONSUMER_SECRET or OAUTH1_ACCESS_TOKEN or OAUTH1_ACCESS_SECRET found in environment variables", file=sys.stderr)


def test_oauth2():
    with unittest.TestCase().assertRaises(LoginError):
        login = wbi_login.OAuth2(consumer_token='wrong', consumer_secret='wrong')
        login.generate_edit_credentials()

    if OAUTH2_CONSUMER_TOKEN and OAUTH2_CONSUMER_SECRET:
        login = wbi_login.OAuth2(consumer_token=OAUTH2_CONSUMER_TOKEN, consumer_secret=OAUTH2_CONSUMER_SECRET)
        login.generate_edit_credentials()
    else:
        print("no OAUTH2_CONSUMER_TOKEN or CLIENT_SECRET found in environment variables", file=sys.stderr)


def test_mismatch_api_url():
    if WDUSER and WDPASS:
        login = wbi_login.Login(user=WDUSER, password=WDPASS)
        login.generate_edit_credentials()
        with pytest.raises(ValueError):
            mediawiki_api_call_helper(data={}, login=login, mediawiki_api_url='https://unsdfdskfjljzkerezr.org/w/api.php')



from collections import defaultdict
from typing import Any

from wikibaseintegrator import WikibaseIntegrator, wbi_fastrun
from wikibaseintegrator.datatypes import BaseDataType, ExternalID, Item
from wikibaseintegrator.wbi_config import config as wbi_config
from wikibaseintegrator.wbi_enums import ActionIfExists

wbi_config['USER_AGENT'] = 'WikibaseIntegrator-pytest/1.0 (test_wbi_fastrun.py)'

wbi = WikibaseIntegrator()


def test_query_data():
    """
    test_fastrun.test_query_data
    This hits live wikidata and may change !!

    This tests that the fast run container correctly queries data from wikidata and stores it in the appropriate format
    without getting references
    """
    frc = wbi_fastrun.FastRunContainer(base_filter=[BaseDataType(prop_nr='P699')], base_data_type=BaseDataType)
    # get a string value
    frc._query_data('P699')
    # wikidata-item value
    frc._query_data('P828')
    # uri value
    frc._query_data('P2888')

    # https://www.wikidata.org/wiki/Q10874
    assert 'Q10874' in frc.prop_data
    assert 'P699' in frc.prop_data['Q10874']
    # the ID may change, so retrieve it
    statement_id = list(frc.prop_data['Q10874']['P699'].keys())[0]
    d = frc.prop_data['Q10874']['P699'][statement_id]
    # d looks like: {'qual': set(), 'ref': {}, 'v': 'DOID:1432'}
    assert all(x in d for x in {'qual', 'ref', 'v'})
    assert frc.prop_data['Q10874']['P699'][statement_id]['v'].startswith('"DOID:')

    # item
    assert list(frc.prop_data['Q10874']['P828'].values())[0]['v'] == "Q18228398"

    # uri
    v = {x['v'] for x in frc.prop_data['Q10874']['P2888'].values()}
    assert all(y.startswith("<http") for y in v)


# # Fail too often
# def test_query_data_unit():
#     """
#     test_fastrun.test_query_data_unit
#     This hits live wikidata and may change !!
#     """
#     frc = wbi_fastrun.FastRunContainer(base_filter={'P2044': '', 'P30': 'Q46', 'P31': 'Q5119'}, base_data_type=BaseDataType)
#     # get a quantity value
#     frc._query_data('P2044', use_units=True)
#
#     assert 'Q90' in frc.prop_data
#     assert 'P2044' in frc.prop_data['Q90']
#     statement_id = list(frc.prop_data['Q90']['P2044'].keys())[0]
#     assert frc.prop_data['Q90']['P2044'][statement_id]['unit'] == 'Q11573'


def test_query_data_ref():
    """
    test_fastrun.test_query_data_ref
    This hits live wikidata and may change !!

    This tests that the fast run container correctly queries data from wikidata and stores it in the appropriate format
    WITH getting references
    """
    frc = wbi_fastrun.FastRunContainer(base_filter=[BaseDataType(prop_nr='P699')], base_data_type=BaseDataType, use_refs=True)
    frc._query_data('P699')

    # https://www.wikidata.org/wiki/Q10874
    assert 'Q10874' in frc.prop_data
    assert 'P699' in frc.prop_data['Q10874']
    # the ID may change, so retrieve it
    statement_id = list(frc.prop_data['Q10874']['P699'].keys())[0]
    d = frc.prop_data['Q10874']['P699'][statement_id]
    # d looks like:
    """
    {'qual': set(),
     'ref': {'16c138dfc51df49853f1a9d79f31e2234853842c': {('P248', 'Q30988716'),
            ('P699', 'DOID:1432'),
            ('P813', '+2017-07-05T00:00:00Z')}},
      'v': 'DOID:1432'}
    """
    assert all(x in d for x in {'qual', 'ref', 'v'})
    assert frc.prop_data['Q10874']['P699'][statement_id]['v'].startswith('"DOID:')
    assert len(d['ref']) > 0
    ref_id = list(d['ref'].keys())[0]
    ref = d['ref'][ref_id]
    assert len(ref) > 1


class FastRunContainerFakeQueryDataEnsembl(wbi_fastrun.FastRunContainer):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.prop_dt_map = {'P248': 'wikibase-item', 'P594': 'external-id'}
        self.prop_data['Q14911732'] = {'P594': {
            'fake statement id': {
                'qual': set(),
                'ref': {'fake ref id': {
                    ('P248',
                     'Q106833387'),
                    ('P594',
                     'ENSG00000123374')}},
                'unit': '1',
                'v': '"ENSG00000123374"'}}}
        self.rev_lookup = defaultdict(set)
        self.rev_lookup['"ENSG00000123374"'].add('Q14911732')


class FastRunContainerFakeQueryDataEnsemblNoRef(wbi_fastrun.FastRunContainer):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.prop_dt_map = {'P248': 'wikibase-item', 'P594': 'external-id'}
        self.prop_data['Q14911732'] = {'P594': {
            'fake statement id': {
                'qual': set(),
                'ref': {},
                'v': 'ENSG00000123374'}}}
        self.rev_lookup = defaultdict(set)
        self.rev_lookup['"ENSG00000123374"'].add('Q14911732')


def test_fastrun_ref_ensembl():
    # fastrun checks refs
    frc = FastRunContainerFakeQueryDataEnsembl(base_filter=[BaseDataType(prop_nr='P594'), Item(prop_nr='P703', value='Q15978631')], base_data_type=BaseDataType, use_refs=True)

    # statement has no ref
    statements = [ExternalID(value='ENSG00000123374', prop_nr='P594')]
    assert frc.write_required(data=statements)

    # statement has the same ref
    statements = [ExternalID(value='ENSG00000123374', prop_nr='P594', references=[[Item("Q106833387", prop_nr="P248"), ExternalID("ENSG00000123374", prop_nr="P594")]])]
    assert not frc.write_required(data=statements)

    # new statement has an different stated in
    statements = [ExternalID(value='ENSG00000123374', prop_nr='P594', references=[[Item("Q99999999999", prop_nr="P248"), ExternalID("ENSG00000123374", prop_nr="P594", )]])]
    assert frc.write_required(data=statements)

    # fastrun don't check references, statement has no reference,
    frc = FastRunContainerFakeQueryDataEnsemblNoRef(base_filter=[BaseDataType(prop_nr='P594'), Item(prop_nr='P703', value='Q15978631')], base_data_type=BaseDataType,
                                                    use_refs=False)
    statements = [ExternalID(value='ENSG00000123374', prop_nr='P594')]
    assert not frc.write_required(data=statements)

    # fastrun don't check references, statement has reference,
    frc = FastRunContainerFakeQueryDataEnsemblNoRef(base_filter=[BaseDataType(prop_nr='P594'), Item(prop_nr='P703', value='Q15978631')], base_data_type=BaseDataType,
                                                    use_refs=False)
    statements = [ExternalID(value='ENSG00000123374', prop_nr='P594', references=[[Item("Q123", prop_nr="P31")]])]
    assert not frc.write_required(data=statements)


class FakeQueryDataAppendProps(wbi_fastrun.FastRunContainer):
    # an item with three values for the same property
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.prop_dt_map = {'P527': 'wikibase-item', 'P248': 'wikibase-item', 'P594': 'external-id'}

        self.rev_lookup = defaultdict(set)
        self.rev_lookup['Q24784025'].add('Q3402672')
        self.rev_lookup['Q24743729'].add('Q3402672')
        self.rev_lookup['Q24782625'].add('Q3402672')

        self.prop_data['Q3402672'] = {'P527': {
            'Q3402672-11BA231B-857B-498B-AC4F-91D71EE007FD': {'qual': set(),
                                                              'ref': {
                                                                  '149c9c7ba4e246d9f09ce3ed0cdf7aa721aad5c8': {
                                                                      ('P248', 'Q3047275'),
                                                                  }},
                                                              'v': 'Q24784025'},
            'Q3402672-15F54AFF-7DCC-4DF6-A32F-73C48619B0B2': {'qual': set(),
                                                              'ref': {
                                                                  '149c9c7ba4e246d9f09ce3ed0cdf7aa721aad5c8': {
                                                                      ('P248', 'Q3047275'),
                                                                  }},
                                                              'v': 'Q24743729'},
            'Q3402672-C8F11D55-1B11-44E5-9EAF-637E062825A4': {'qual': set(),
                                                              'ref': {
                                                                  '149c9c7ba4e246d9f09ce3ed0cdf7aa721aad5c8': {
                                                                      ('P248', 'Q3047275')}},
                                                              'v': 'Q24782625'}}}


def test_append_props():
    qid = 'Q3402672'
    # https://www.wikidata.org/wiki/Q3402672#P527

    # don't consider refs
    frc = FakeQueryDataAppendProps(base_filter=[BaseDataType(prop_nr='P352'), Item(prop_nr='P703', value='Q15978631')], base_data_type=BaseDataType)
    # with append
    statements = [Item(value='Q24784025', prop_nr='P527')]
    assert frc.write_required(data=statements, action_if_exists=ActionIfExists.APPEND_OR_REPLACE, cqid=qid) is False
    # with force append
    statements = [Item(value='Q24784025', prop_nr='P527')]
    assert frc.write_required(data=statements, action_if_exists=ActionIfExists.FORCE_APPEND, cqid=qid) is True
    # without append
    statements = [Item(value='Q24784025', prop_nr='P527')]
    assert frc.write_required(data=statements, cqid=qid) is True

    # if we are in append mode, and the refs are different, we should write
    frc = FakeQueryDataAppendProps(base_filter=[BaseDataType(prop_nr='P352'), Item(prop_nr='P703', value='Q15978631')], base_data_type=BaseDataType, use_refs=True)
    # with append
    statements = [Item(value='Q24784025', prop_nr='P527')]
    assert frc.write_required(data=statements, cqid=qid, action_if_exists=ActionIfExists.APPEND_OR_REPLACE) is True
    # without append
    statements = [Item(value='Q24784025', prop_nr='P527')]
    assert frc.write_required(data=statements, cqid=qid) is True



from setuptools import setup

# Metadata goes in setup.cfg. These are here for GitHub's dependency graph.
# https://github.com/pallets/flask/blob/main/setup.py
setup(
    name="wikibaseintegrator",
    install_requires=[
        "backoff >= 1.11.1,< 2.2.0",
        "mwoauth ~= 0.3.8",
        "oauthlib ~= 3.2.0",
        "requests >= 2.27.1,< 2.29.0",
        "requests-oauthlib ~= 1.3.1",
        "ujson ~= 5.4.0"
    ],
    extras_require={
        "dev": [
            "pytest",
            "pylint",
            "pylint-exit",
            "mypy",
            "codespell",
            "flynt"
        ],
        "docs": [
            "Sphinx ~= 4.5.0",
            "readthedocs-sphinx-ext ~= 2.1.5",
            "sphinx-rtd-theme ~= 1.0.0",
            "sphinx_github_changelog ~= 1.2.0",
            "m2r2 ~= 0.3.2",
            "sphinx-autodoc-typehints ~= 1.18.1"
        ],
        "notebooks": ["jupyter"],
        "coverage": ["pytest-cov"],
    },
)



import unittest

import requests

from wikibaseintegrator import WikibaseIntegrator
from wikibaseintegrator.datatypes import BaseDataType, Item
from wikibaseintegrator.wbi_config import config as wbi_config
from wikibaseintegrator.wbi_exceptions import NonExistentEntityError

wbi_config['USER_AGENT'] = 'WikibaseIntegrator-pytest/1.0 (test_entity_item.py)'

wbi = WikibaseIntegrator()


class TestEntityItem(unittest.TestCase):

    def test_get(self):
        # Test with complete id
        assert wbi.item.get('Q582').id == 'Q582'
        # Test with numeric id as string
        assert wbi.item.get('582').id == 'Q582'
        # Test with numeric id as int
        assert wbi.item.get(582).id == 'Q582'

        # Test with invalid id
        with self.assertRaises(ValueError):
            wbi.item.get('L5')

        # Test with zero id
        with self.assertRaises(ValueError):
            wbi.item.get(0)

        # Test with negative id
        with self.assertRaises(ValueError):
            wbi.item.get(-1)

        # Test with negative id
        with self.assertRaises(NonExistentEntityError):
            wbi.item.get("Q99999999999999")

    def test_get_json(self):
        assert wbi.item.get('Q582').get_json()['labels']['fr']['value'] == 'Villeurbanne'

    def test_write(self):
        with self.assertRaises(requests.exceptions.JSONDecodeError):
            wbi.item.get('Q582').write(allow_anonymous=True, mediawiki_api_url='https://httpstat.us/200')

    def test_write_not_required(self):
        assert not wbi.item.get('Q582').write_required(base_filter=[BaseDataType(prop_nr='P1791')])

    def test_write_required(self):
        item = wbi.item.get('Q582')
        item.claims.add(Item(prop_nr='P1791', value='Q42'))
        assert item.write_required([BaseDataType(prop_nr='P1791')])

    def test_write_not_required_ref(self):
        assert not wbi.item.get('Q582').write_required(base_filter=[BaseDataType(prop_nr='P2581')], use_refs=True)

    def test_write_required_ref(self):
        item = wbi.item.get('Q582')
        item.claims.get('P2581')[0].references.references.pop()
        assert item.write_required(base_filter=[BaseDataType(prop_nr='P2581')], use_refs=True)

    def test_long_item_id(self):
        assert wbi.item.get('Item:Q582').id == 'Q582'



"""
WikibaseIntegrator Library
~~~~~~~~~~~~~~~~~~~~~

WikibaseIntegrator is a Wikibase library, written in Python, for human beings.
Basic read usage:

   >>> from wikibaseintegrator import WikibaseIntegrator
   >>> from wikibaseintegrator.wbi_config import config
   >>> config['USER_AGENT'] = 'Item Get Notebook'
   >>> wbi = WikibaseIntegrator()
   >>> q42 = wbi.item.get('Q42')
   >>> q42.labels.get('en').value
   'Douglas Adams'

Full documentation is available at <https://wikibaseintegrator.readthedocs.io/>.
"""
import pkg_resources

from .wikibaseintegrator import WikibaseIntegrator

try:
    __version__ = pkg_resources.get_distribution('wikibaseintegrator').version
except pkg_resources.DistributionNotFound as e:  # pragma: no cover
    __version__ = 'dev'
